import path from 'node:path';
import { fileURLToPath } from 'node:url';
import { spawn } from 'node:child_process';

function createPromise() {
	let promise, resolve, reject;

	promise = new Promise((a, b) => {
		resolve = a;
		reject = b;
	});

	return {promise, resolve, reject}
}

const __dirname = path.dirname(
	fileURLToPath(import.meta.url)
);

function createNodeWorkerProcess(options) {
	let node_binary_path = "node", silent = true;

	if ("node_binary" in options) {
		node_binary_path = options.node_binary;
	}

	if ("silent" in options) {
		silent = options.silent;
	}

	const stdio = silent ? ["pipe", "pipe", "pipe", "ipc"] : ["pipe", "inherit", "inherit", "ipc"];

	const child = spawn(node_binary_path, [
		path.resolve(__dirname, "bootstrap.mjs")
	], {
		stdio
	});

	return child
}

function createWorkerInstance({
	child,
	worker_message_buffer
}) {
	let instance = {};

	let currentOnMessageHandler = null;

	child.on("message", msg => {
		if (typeof currentOnMessageHandler === "function") {
			currentOnMessageHandler(msg);
		}
	});

	Object.defineProperty(instance, "onMessage", {
		set(handler) {
			// dump message buffer first
			while (worker_message_buffer.length) {
				const msg = worker_message_buffer.shift();

				handler(msg);
			}

			currentOnMessageHandler = handler;
		}
	});

	Object.defineProperty(instance, "sendMessage", {
		enumerable: true,
		get() { return (str) => child.send(str) }
	});

	Object.defineProperty(instance, "terminate", {
		enumerable: true,
		get() { return () => child.kill("SIGKILL") }
	});

	return instance
}

function nodeCreateWorker(worker_file_path, worker_args, additional = {}) {
	let {promise, resolve, reject} = createPromise();

	const init_token = Math.random().toString(32) + "_" + Math.random().toString(32);

	let child = createNodeWorkerProcess(additional);
	let worker_message_buffer = [];

	child.on("error", reject);

	child.on("message", msg => {
		if (msg === init_token) {
			resolve(
				createWorkerInstance({
					child,
					worker_message_buffer
				})
			);
		}
		// buffer other messages between worker and main script
		else {
			worker_message_buffer.push(msg);
		}
	});

	child.send("init" + JSON.stringify({
		worker_file_path,
		worker_args,
		init_token,
		additional
	}));

	return promise
}

export { nodeCreateWorker as default };
